// -*- tab-width: 4; indent-tabs-mode: nil; c-basic-offset: 2 -*-
// vi: set et ts=4 sw=2 sts=2:
// SPDX-FileCopyrightInfo: Copyright © DUNE Project contributors, see file LICENSE.md in module root
// SPDX-License-Identifier: LicenseRef-GPL-2.0-only-with-DUNE-exception
#ifndef DUNE_LOCALFUNCTIONS_ENRICHED_HH
#define DUNE_LOCALFUNCTIONS_ENRICHED_HH

#include <dune/localfunctions/enriched/cubeq1bubble.hh>
#include <dune/localfunctions/enriched/simplexp1bubble.hh>

#endif // DUNE_LOCALFUNCTIONS_ENRICHED_HH
